// A program to display a parametric surface
//
// Use '1' to toggle visibility of the surface

#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glut.h>
#include "Trackball.h"
#include "Lighting.h"
#include "Geometry.h"

const int windowWidth=400;
const int windowHeight=400;

bool SURFACE_ENABLED=true;


// material properties 
GLfloat mat_specular[4];
GLfloat mat_ambient_and_diffuse1[4];	// colour for the original surface
GLfloat mat_ambient_and_diffuse2[4];	// colour for bilinear interpolant
GLfloat mat_shininess[1];

CTrackball trackball;
CLighting lighting;

// sample points
const int NUM_SAMPLES=5;
float p[NUM_SAMPLES][NUM_SAMPLES];

// original surface and derivatives
inline float f(float x, float y)
{
	return -0.5*x*y+0.2*exp(x)-0.1*sin(2*Pi*y);
}

inline float dfdx(float x, float y)
{
	return -0.5*y+0.2*exp(x);
}

inline float dfdy(float x, float y)
{
	return -0.5*x-0.1*cos(Pi*y)*2*Pi;
}



void handleMouseMotion(int x, int y)
{	
	trackball.tbMotion(x, y);
}

void handleMouseClick(int button, int state, int x, int y)
{
	trackball.tbMouse(button, state, x, y);
}

void handleKeyboardEvent(unsigned char key, int x, int y)
{
	switch (key)
	{
		case '1': SURFACE_ENABLED=!SURFACE_ENABLED;
					glutPostRedisplay(); break;
		default: trackball.tbKeyboard(key);
	}
}

void display(void)
{
	glMatrixMode( GL_MODELVIEW );	// Set the view matrix ...
	glLoadIdentity();				// ... to identity.
    gluLookAt(0,0,4, 0,0,0, 0,1,0); // camera is on the z-axis
	trackball.tbMatrix();			// rotate scene using the trackball
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glTranslatef(-0.5,-0.5,0);
	int i,j;

	// draw the ground mesh
	glColor3f (0.0, 0.0, 0.5);		
	glBegin(GL_LINES);
	float c;
	for(i=0;i<11;i++){
		c=((float) i)/10.0f;
		glVertex3f(c,0,0);
		glVertex3f(c,1,0);
		glVertex3f(0,c,0);
		glVertex3f(1,c,0);
	}
	glEnd();


	glEnable(GL_NORMALIZE);
	glFrontFace(GL_CW);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	lighting.enable();
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

	if (SURFACE_ENABLED){
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_ambient_and_diffuse1);
		int numSegments=20;
		float x,y;
		CVec3df v1,v2,n;
		for(i=0;i<numSegments;i++)
		{
			glBegin(GL_QUAD_STRIP);
			for(j=0;j<=numSegments;j++)
			{	
				y=(float) j/(float) numSegments;
				x=(float) i/(float) numSegments;
				v1.setVector(1,0,dfdx(x,y));	
				v2.setVector(0,1,dfdy(x,y));	
				n=cross(v2,v1); 
				n.normaliseDestructive();
				glNormal3fv(n.getArray());
				glVertex3f(x,y,f(x,y));
				x=(float) (i+1)/(float) numSegments;
				v1.setVector(1,0,dfdx(x,y));	
				v2.setVector(0,1,dfdy(x,y));	
				n=cross(v2,v1); 
				n.normaliseDestructive();
				glNormal3fv(n.getArray());
				glVertex3f(x,y,f(x,y));
			}	
			glEnd();
		}
	}

	lighting.disable();
	glFlush ();
	glutSwapBuffers();
}

void init(void) 
{
	// select clearing color (for glClear)
	glClearColor (1,1,1,1);	// RGB-value for black
	// enable depth buffering
	glEnable(GL_DEPTH_TEST);
	// initialize view (simple orthographic projection)
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(33,1,2,6);
	trackball.tbInit(GLUT_LEFT_BUTTON);

	// material properties 
	mat_ambient_and_diffuse1[0]=0.8;		// red material ...
	mat_ambient_and_diffuse1[1]=0.1;
	mat_ambient_and_diffuse1[2]=0.1;
	mat_ambient_and_diffuse1[3]=1;
	mat_ambient_and_diffuse2[0]=0.1;		// blue material ...
	mat_ambient_and_diffuse2[1]=0.1;
	mat_ambient_and_diffuse2[2]=0.8;
	mat_ambient_and_diffuse2[3]=1;
	mat_specular[0]=0.8f;				// ... with white highlights
	mat_specular[1]=0.8f;				// if light source is reflected
	mat_specular[2]=0.8f;				// on the material surface.
	mat_specular[3]=1;
	mat_shininess[0]=100;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);

	glShadeModel(GL_SMOOTH);
}

void reshape(int width, int height ) {
	// Called at start, and whenever user resizes component
	int size = min(width, height);
	glViewport(0, 0, size, size);  // Largest possible square
	trackball.tbReshape(width, height);
}


// create a double buffered colour window
int main(int argc, char** argv)
{
	glutInit(&argc, argv);		
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(windowWidth, windowHeight); 
	glutInitWindowPosition(100, 100);
	glutCreateWindow("Parametric Surface");
	init ();								// initialise view
	glutMouseFunc(handleMouseClick);		// Set function to handle mouse clicks
	glutMotionFunc(handleMouseMotion);		// Set function to handle mouse motion
	glutKeyboardFunc(handleKeyboardEvent);	// Set function to handle keyboard input
	glutDisplayFunc(display);		// Set function to draw scene
	glutReshapeFunc(reshape);		// Set function called if window gets resized
	glutMainLoop();
	return 0;
}
